# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
from __future__ import annotations

import inspect
import os
import sys

import abqcy

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "abqcy"
copyright = "2023, WANG Hailin"
author = "WANG Hailin"
release = version = abqcy.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

sys.path.insert(0, os.path.abspath(".."))
extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.linkcode",
    "sphinx.ext.githubpages",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
    "sphinx_codeautolink",
    "sphinx_design",
    "sphinx_inline_tabs",
    "sphinxcontrib.programoutput",
    "hoverxref.extension",
    "autoapi.extension",
]

# MyST configuration
myst_enable_extensions = [
    "colon_fence",
]

# AutoAPI configuration
autoapi_dirs = ["../abqcy"]
autoapi_ignore = ["*_version.py", "*__main__.py"]
autoapi_options = [
    "members",
    "undoc-members",
    "private-members",
    "show-inheritance",
    "show-module-summary",
    "special-members",
    # 'imported-members',
]
autoapi_template_dir = "_autoapi_templates"

# Default behavior for code block concatenation for sphinx_codeautolink
codeautolink_concat_default = False

# Suppress warnings
suppress_warnings = [
    "app.add_directive",
]

intersphinx_mapping = {
    "jinjia2": ("https://jinja.palletsprojects.com/en/3.0.x/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "pytest": ("https://pytest.org/en/stable/", None),
    "python": ("https://docs.python.org/3/", None),
    "readthedocs": ("https://docs.readthedocs.io/en/stable/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
}

# Hoverxref configuration
hoverxref_auto_ref = True
hoverxref_domains = ["py"]
hoverxref_roles = [
    "numref",
    "confval",
    "setting",
    "option",
    "doc",  # Documentation pages
    "term",  # Glossary terms
]
hoverxref_role_types = {
    "doc": "modal",  # for whole docs
    "mod": "modal",  # for Python Sphinx Domain
    "class": "tooltip",  # for Python Sphinx Domain
    "func": "tooltip",  # for Python Sphinx Domain
    "meth": "tooltip",  # for Python Sphinx Domain
    "attr": "tooltip",  # for Python Sphinx Domain
    "exc": "tooltip",  # for Python Sphinx Domain
    "obj": "tooltip",  # for Python Sphinx Domain
    "ref": "tooltip",  # for hoverxref_auto_ref config
    "confval": "tooltip",  # for custom object
    "term": "tooltip",  # for glossaries
    "numref": "tooltip",
}
hoverxref_intersphinx = [
    "numpy",
    "pytest",
    "python",
    "readthedocs",
]

# Show short type hints for user-defined classes and defaults for parameters
add_module_names = False
autodoc_default_options = {
    "undoc-members": False,
}
autodoc_typehints_format = "short"
autoclass_content = "both"
typehints_defaults = "comma"
typehints_document_rtype = False
python_use_unqualified_type_names = True

# Figure numbering
numfig = True

# True to convert the type definitions in the docstrings as references. Defaults to False.
napoleon_preprocess_types = True

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]


# linkcode source
def linkcode_resolve(domain: str, info: dict[str, str | list[str]]):
    """Resolve linkcode source

    Parameters
    ----------
    domain : str
        specifies the language domain the object is in
    info : dict[str, str | list[str]]
        a dictionary with the following keys guaranteed to be present (dependent on the domain)

        - py: module (name of the module), fullname (name of the object)
        - c: names (list of names for the object)
        - cpp: names (list of names for the object)
        - javascript: object (name of the object), fullname (name of the item)

    Returns
    -------
    source url of the object
    """
    if domain != "py":
        return None

    modname = info["module"]
    fullname = info["fullname"]

    filename = modname.replace(".", "/")
    baseurl = f"https://github.com/haiiliin/abqcy/blob/main/{filename}.py"

    submod = sys.modules.get(modname)
    if submod is None:
        return baseurl

    obj = submod
    for part in fullname.split("."):
        try:
            obj = getattr(obj, part)
        except Exception:
            return baseurl
    try:
        source, lineno = inspect.getsourcelines(obj)
    except Exception:
        return baseurl

    return baseurl + f"#L{lineno}-L{lineno + len(source) - 1}"



STANDARD = [
    "CREEP",
    "DFLOW",
    "DFLUX",
    "DISP",
    "DLOAD",
    "FILM",
    "FLOW",
    "FRIC",
    "FRIC_COEF",
    "GAPCON",
    "GAPELECTR",
    "HARDINI",
    "HETVAL",
    "MPC",
    "ORIENT",
    "RSURFU",
    "SDVINI",
    "SIGINI",
    "UAMP",
    "UANISOHYPER_INV",
    "UANISOHYPER_STRAIN",
    "UCORR",
    "UCREEPNETWORK",
    "UDAMAGEMF",
    "UDECURRENT",
    "UDEMPOTENTIAL",
    "UDMGINI",
    "UDSECURRENT",
    "UEL",
    "UELEMDRESP",
    "UELMAT",
    "UEPACTIVATIONFACET",
    "UEPACTIVATIONSETUP",
    "UEPACTIVATIONVOL",
    "UEXPAN",
    "UEXTERNALDB",
    "UFIELD",
    "UFLUID",
    "UFLUIDCONNECTORLOSS",
    "UFLUIDCONNECTORVALVE",
    "UFLUIDLEAKOFF",
    "UFLUIDPIPEFRICTION",
    "UGENS",
    "UHARD",
    "UHYPEL",
    "UHYPER",
    "UHYPER_STRETCH",
    "UINTER",
    "UMASFL",
    "UMAT",
    "UMATHT",
    "UMDFLUX",
    "UMDFLUXSETUP",
    "UMESHMOTION",
    "UMIXMODEFATIGUE",
    "UMOTION",
    "UMULLINS",
    "UPOREP",
    "UPRESS",
    "UPSD",
    "URDFIL",
    "USDFLD",
    "USUPERELASHARDMOD",
    "UTEMP",
    "UTRACLOAD",
    "UTRS",
    "UTRSNETWORK",
    "UVARM",
    "UWAVE",
    "UXFEMNONLOCALWEIGHT",
    "VOIDRI",
]

EXPLICIT = [
    "VDFLUX",
    "VDISP",
    "VDLOAD",
    "VEXTERNALDB",
    "VFABRIC",
    "VFRIC",
    "VFRIC_COEF",
    "VFRICTION",
    "VHETVAL",
    "VUAMP",
    "VUANISOHYPER_INV",
    "VUANISOHYPER_STRAIN",
    "VUCHARLENGTH",
    "VUCREEPNETWORK",
    "VUEL",
    "VUEOS",
    "VUEXPAN",
    "VUFIELD",
    "VUFLUIDEXCH",
    "VUFLUIDEXCHEFFAREA",
    "VUHARD",
    "VUINTER",
    "VUINTERACTION",
    "VUMAT",
    "VUMATHT",
    "VUMULLINS",
    "VUSDFLD",
    "VUSUPERELASHARDMOD",
    "VUTRS",
    "VUVISCOSITY",
    "VWAVE",
]
subs = STANDARD + EXPLICIT
subs += [sub.lower() for sub in subs]



cdef extern from "<aba_for_c.h>":
    pass


cdef extern void umat(
    double *stress, double *statev, double *ddsdde, double *sse, double *spd, double *scd,
    double *rpl, double *ddsddt, double *drplde, double *drpldt, double *stran, double *dstran,
    double *time, double *dtime, double *temp, double *dtemp, double *predef, double *dpred,
    char *cmname, int *ndi, int *nshr, int *ntens, int *nstatv, double *props, int *nprops,
    double *coords, double *drot, double *pnewdt, double *celent, double *dfgrd0,
    double *dfgrd1, int *noel, int *npt, int *layer, int *kspt, int *jstep, int *kinc,
)



import fire

from abqcy.cli import AbqcyCLI


def main():
    fire.core.Display = lambda lines, out: out.write("\n".join(lines) + "\n")
    fire.Fire(AbqcyCLI)


if __name__ == "__main__":
    main()



def umat(
    stress, statev, ddsdde, sse, spd, scd, rpl, ddsddt, drplde, drpldt, stran, dstran,
    time, dtime, temp, dtemp, predef, dpred, cmname, ndi, nshr, ntens, nstatv, props,
    nprops, coords, drot, pnewdt, celent, dfgrd0, dfgrd1, noel, npt, layer, kspt,
    jstep, kinc,
):  # fmt: skip
    r"""Abaqus UMAT interface for the material model.

    Parameters
    ----------
    stress : double[:], shape (ntens,), defined in all situations
        This array is passed in as the stress tensor at the beginning of the increment and must be updated in this
        routine to be the stress tensor at the end of the increment. If you specified initial stresses Initial
        Conditions, this array will contain the initial stresses at the start of the analysis. The size of this
        array depends on the value of ``NTENS`` as defined below. In finite-strain problems the stress tensor has
        already been rotated to account for rigid body motion in the increment before ``UMAT`` is called, so that
        only the corotational part of the stress integration should be done in ``UMAT``. The measure of stress used
        is ``true`` (Cauchy) stress.

        If the ``UMAT`` utilizes a hybrid formulation that is total (as opposed to the default incremental behavior),
        the stress array is extended beyond NTENS. The first NTENS entries of the array contain the stresses, as
        described above. The additional quantities are as follows::

            STRESS(NTENS+1)

        Read only :math:`\hat{J}`::

            STRESS(NTENS+2)

        Write only :math:`\widehat{K}=J \frac{\partial^2 U}{\partial \widehat{J}^2}`, and::

            STRESS(NTENS+3)

        Write only
        :math:`\frac{\partial \widehat{K}}{\partial \widehat{J}}=J \frac{\partial^3 U}{\partial \widehat{J}^3}`,
        where :math:`U` is the volumetric part of the strain energy density potential.
    statev : double[:], shape (nstatv,), defined in all situations
        An array containing the solution-dependent state variables. These are passed in as the values at the beginning
        of the increment unless they are updated in user subroutines ``USDFLD`` or ``UEXPAN``, in which case the
        updated values are passed in. In all cases ``STATEV`` must be returned as the values at the end of the
        increment. The size of the array is defined as described in Allocating Space for Solution-Dependent State
        Variables.

        In finite-strain problems any vector-valued or tensor-valued state variables must be rotated to account for
        rigid body motion of the material, in addition to any update in the values associated with constitutive
        behavior. The rotation increment matrix, ``DROT``, is provided for this purpose.
    ddsdde : double[:,:], shape (ntens, ntens), defined in all situations
        Jacobian matrix of the constitutive model,
        :math:`\mathbf{C}=(1 / J) \partial \Delta(J \boldsymbol{\sigma}) / \partial \Delta \boldsymbol{\varepsilon}`,
        where :math:`\Delta(J \boldsymbol{\sigma})` are the Kirchhoff stress increments, :math:`J` is the determinant
        of the deformation gradient representing the volume change from the reference configuration, :math:`\sigma` is
        the Cauchy  stress, and :math:`\Delta \varepsilon` are the strain increments. If the volume change is small,
        the Jacobian matrix can be approximated as
        :math:`\partial \Delta \boldsymbol{\sigma} / \partial \Delta \varepsilon`, where :math:`\Delta \sigma` are the
        Cauchy stress increments. ``DDSDDE(I,J)`` defines the change in the Ith stress component at the end of the time
        increment caused by an infinitesimal perturbation of the Jth component of the strain increment array. Unless
        you invoke the unsymmetric equation solution capability for the user-defined material, Abaqus/Standard will use
        only the symmetric part of ``DDSDDE``. The symmetric part of the matrix is calculated by taking one half the sum
        of the matrix and its transpose.

        For viscoelastic behavior in the frequency domain, the Jacobian matrix must be dimensioned as
        ``DDSDDE(NTENS,NTENS,2)``. The stiffness contribution (storage modulus) must be provided in
        ``DDSDDE(NTENS,NTENS,1)``, while the damping contribution (loss modulus) must be provided in
        ``DDSDDE(NTENS,NTENS,2)``.
    sse, spd, scd : double, defined in all situations
        Specific elastic strain energy, plastic dissipation, and ``creep`` dissipation, respectively. These are passed in
        as the values at the start of the increment and should be updated to the corresponding specific energy values
        at the end of the increment. They have no effect on the solution, except that they are used for energy output.
    rpl : double, defined only in a fully coupled thermal-stress or a coupled thermal-electrical-structural analysis
        Volumetric heat generation per unit time at the end of the increment caused by mechanical working of the
        material.
    ddsddt : double[:,:], shape (ntens, ntens), defined only in a fully coupled thermal-stress or a coupled thermal-electrical-structural analysis
        Variation of the stress increments with respect to the temperature.
    drplde : double[:], shape (ntens,), defined only in a fully coupled thermal-stress or a coupled thermal-electrical-structural analysis
        Variation of ``RPL`` with respect to the strain increments.
    drpldt : double, defined only in a fully coupled thermal-stress or a coupled thermal-electrical-structural analysis
        Variation of ``RPL`` with respect to the strain increments.
    stran : double[:], shape (ntens,), passed for information only
        An array containing the total strains at the beginning of the increment. If thermal expansion is included in
        the same material definition, the strains passed into UMAT are the mechanical strains only (that is, the
        thermal strains computed based upon the thermal expansion coefficient have been subtracted from the total
        strains). These strains are available for output as the ``elastic`` strains.

        In finite-strain problems the strain components have been rotated to account for rigid body motion in the
        increment before UMAT is called and are approximations to logarithmic strain.
    dstran : double[:], shape (ntens,), passed for information only
        Array of strain increments. If thermal expansion is included in the same material definition, these are the
        mechanical strain increments (the total strain increments minus the thermal strain increments).
    time : double, passed for information only
        - ``time(1)``: Value of step time at the beginning of the current increment or frequency.
        - ``time(2)``: Value of step time at the beginning of the current increment or frequency.
    dtime : double, passed for information only
        Value of step time at the beginning of the current increment or frequency.
    temp : double, passed for information only
        Temperature at the beginning of the increment.
    dtemp : double, passed for information only
        Increment of temperature.
    predef : double[:], shape (npr-ed,), passed for information only
        Array of interpolated values of predefined field variables at this point at the start of the increment, based
        on the values read in at the nodes.
    dpred : double[:], shape (npred,), passed for information only
        Array of increments of predefined field variables.
    cmname : str, passed for information only
        User-defined material name, left justified. Some internal material models are given names starting with the
        ``ABQ_`` character string. To avoid conflict, you should not use ``ABQ_`` as the leading string for ``CMNAME``.
    ndi : int, passed for information only
        Number of direct stress components at this point.
    nshr : int, passed for information only
        Number of engineering shear stress components at this point.
    ntens : int, passed for information only
        Size of the stress or strain component array (``NDI`` + ``NSHR``).
    nstatv : int, passed for information only
        Number of solution-dependent state variables that are associated with this material type (defined as described
        in Allocating Space for Solution-Dependent State Variables).
    props : double[:], shape (nprops,), passed for information only
        User-specified array of material constants associated with this user material.
    nprops : int, passed for information only
        User-specified array of material constants associated with this user material.
    coords : double[:], shape (3,), passed for information only
        An array containing the coordinates of this point. These are the current coordinates if geometric nonlinearity
        is accounted for during the step (see Defining an Analysis); otherwise, the array contains the original
        coordinates of the point.
    drot : double[:,:], shape (3, 3), passed for information only
        Rotation increment matrix. This matrix represents the increment of rigid body rotation of the basis system in
        which the components of stress (``STRESS``) and strain (``STRAN``) are stored. It is provided so that vector-
        or tensor-valued state variables can be rotated appropriately in this subroutine: stress and strain components
        are already rotated by this amount before ``UMAT`` is called. This matrix is passed in as a unit matrix for
        small-displacement analysis and for large-displacement analysis if the basis system for the material point
        rotates with the material (as in a shell element or when a local orientation is used).
    pnewdt : double, can be updated in any situations
        Ratio of suggested new time increment to the time increment being used (``DTIME``, see discussion later in this
        section). This variable allows you to provide input to the automatic time incrementation algorithms in
        Abaqus/Standard (if automatic time incrementation is chosen). For a quasi-static procedure the automatic time
        stepping that Abaqus/Standard uses, which is based on techniques for integrating standard creep laws (see
        Quasi-Static Analysis), cannot be controlled from within the UMAT subroutine.

        ``PNEWDT`` is set to a large value before each call to ``UMAT``.

        - If ``PNEWDT`` is redefined to be less than 1.0, Abaqus/Standard must abandon the time increment and attempt
          it again with a smaller time increment. The suggested new time increment provided to the automatic time
          integration algorithms is ``PNEWDT`` x ``DTIME``, where the ``PNEWDT`` used is the minimum value for all
          calls to user subroutines that allow redefinition of PNEWDT for this iteration.
        - If ``PNEWDT`` is given a value that is greater than 1.0 for all calls to user subroutines for this iteration
          and the increment converges in this iteration, Abaqus/Standard may increase the time increment. The suggested
          new time increment provided to the automatic time integration algorithms is ``PNEWDT`` x ``DTIME``, where the
          ``PNEWDT`` used is the minimum value for all calls to user subroutines for this iteration.
        - If automatic time incrementation is not selected in the analysis procedure, values of ``PNEWDT`` that are
          greater than 1.0 will be ignored and values of ``PNEWDT`` that are less than 1.0 will cause the job to
          terminate.
    celent : double, passed for information only
        Characteristic element length, which is a typical length of a line across an element for a first-order element;
        it is half of the same typical length for a second-order element. For beams and trusses it is a characteristic
        length along the element axis. For membranes and shells it is a characteristic length in the reference surface.
        For axisymmetric elements it is a characteristic length in the :math:`(r,z)` plane only. For cohesive elements
        it is equal to the constitutive thickness.
    dfgrd0 : double[:,:], shape (3, 3), passed for information only
        Array containing the deformation gradient at the beginning of the increment. If a local orientation is defined
        at the material point, the deformation gradient components are expressed in the local coordinate system defined
        by the orientation at the beginning of the increment. For a discussion regarding the availability of the
        deformation gradient for various element types, see Deformation Gradient.
    dfgrd1 : double[:,:], shape (3, 3), passed for information only
        Array containing the deformation gradient at the end of the increment. If a local orientation is defined at the
        material point, the deformation gradient components are expressed in the local coordinate system defined by the
        orientation. This array is set to the identity matrix if nonlinear geometric effects are not included in the
        step definition associated with this increment. For a discussion regarding the availability of the deformation
        gradient for various element types, see Deformation Gradient.
    noel : int, passed for information only
        Element number.
    npt : int, passed for information only
        Integration point number.
    layer : int, passed for information only
        Layer number (for composite shells and layered solids).
    kspt : int, passed for information only
        Section point number within the current layer.
    jstep : int, passed for information only
        - ``jstep(1)``: Step number.
        - ``jstep(2)``: Procedure type key (see Results File).
        - ``jstep(3)``: 1 if NLGEOM=YES for the current step; 0 otherwise.
        - ``jstep(4)``: 1 if NLGEOM=YES for the current step; 0 otherwise.
    kinc : int, passed for information only
        Increment number.
    """



cdef extern from "<aba_for_c.h>":
    pass


cdef extern void umat(
    double *stress, double *statev, double *ddsdde, double *sse, double *spd, double *scd,
    double *rpl, double *ddsddt, double *drplde, double *drpldt, double *stran, double *dstran,
    double *time, double *dtime, double *temp, double *dtemp, double *predef, double *dpred,
    char *cmname, int *ndi, int *nshr, int *ntens, int *nstatv, double *props, int *nprops,
    double *coords, double *drot, double *pnewdt, double *celent, double *dfgrd0,
    double *dfgrd1, int *noel, int *npt, int *layer, int *kspt, int *jstep, int *kinc,
):
    cdef double E, nu, lam, G
    E, nu = props[0], props[1]
    lam = E * nu / ((1.0 + nu) * (1.0 - 2.0 * nu))
    G = E / (2.0 * (1.0 + nu))

    cdef int i, j
    for i in range(3):
        for j in range(3):
            ddsdde[6 * i + j] = lam
        ddsdde[6 * i + i] += 2.0 * G
        ddsdde[6 * (i + 3) + (i + 3)] = G
    for i in range(6):
        for j in range(6):
            stress[i] += ddsdde[6 * i + j] * dstran[j]



import cython


def umat(
    stress, statev, ddsdde, sse, spd, scd, rpl, ddsddt, drplde, drpldt, stran, dstran,
    time, dtime, temp, dtemp, predef, dpred, cmname, ndi, nshr, ntens, nstatv, props,
    nprops, coords, drot, pnewdt, celent, dfgrd0, dfgrd1, noel, npt, layer, kspt,
    jstep, kinc,
):  # fmt: skip
    cython.declare(E=cython.double, nu=cython.double, lam=cython.double, G=cython.double)
    E, nu = props[0], props[1]
    lam = E * nu / ((1.0 + nu) * (1.0 - 2.0 * nu))
    G = E / (2.0 * (1.0 + nu))

    cython.declare(i=cython.int, j=cython.int)
    for i in range(3):
        for j in range(3):
            ddsdde[6 * i + j] = lam
        ddsdde[6 * i + i] += 2.0 * G
        ddsdde[6 * (i + 3) + (i + 3)] = G
    for i in range(6):
        for j in range(6):
            stress[i] += ddsdde[6 * i + j] * dstran[j]



from __future__ import annotations

import os
import re
import shutil
from pathlib import Path

from abqpy.cli import abaqus
from Cython.Build import cythonize

from abqcy.subs import subs


class AbqcyCLI:
    """The ``abqcy`` command-line interface."""

    def compile(
        self,
        script: str,
        *,
        exclude: list = None,
        nthreads: int = 0,
        aliases: dict = None,
        quiet: bool = False,
        force: bool = False,
        language: str = None,
        exclude_failures: bool = False,
        annotate: bool = True,
        **kwargs,
    ):
        """Compile a Cython script to an Abaqus user subroutine as an object file.

        Parameters
        ----------
        script : str
            The path to the Cython script to compile.
        exclude : list, optional
            When passing glob patterns as ``script``, you can exclude certain
            module names explicitly by passing them into the ``exclude`` option.
        nthreads : int, optional
            The number of concurrent builds for parallel compilation
            (requires the ``multiprocessing`` module).
        aliases : dict, optional
            If you want to use compiler directives like ``# distutils: ...`` but
            can only know at compile time (when running the ``setup.py``) which values
            to use, you can use aliases and pass a dictionary mapping those aliases
            to Python strings when calling :func:`cythonize`. As an example, say you
            want to use the compiler
            directive ``# distutils: include_dirs = ../static_libs/include/``
            but this path isn't always fixed and you want to find it when running
            the ``setup.py``. You can then do ``# distutils: include_dirs = MY_HEADERS``,
            find the value of ``MY_HEADERS`` in the ``setup.py``, put it in a python
            variable called ``foo`` as a string, and then call
            ``cythonize(..., aliases={'MY_HEADERS': foo})``.
        quiet : bool, optional
            If True, Cython won't print error, warning, or status messages during the
            compilation.
        force : bool, optional
            Forces the recompilation of the Cython modules, even if the timestamps
            don't indicate that a recompilation is necessary.
        language : str, optional
            To globally enable C++ mode, you can pass ``language='c++'``. Otherwise, this
            will be determined at a per-file level based on compiler directives.  This
            affects only modules found based on file names.  Extension instances passed
            into :func:`cythonize` will not be changed. It is recommended to rather
            use the compiler directive ``# distutils: language = c++`` than this option.
        exclude_failures : bool, optional
            For a broad 'try to compile' mode that ignores compilation
            failures and simply excludes the failed extensions,
            pass ``exclude_failures=True``. Note that this only
            really makes sense for compiling ``.py`` files which can also
            be used without compilation.
        annotate : bool, optional
            Whether to generate an HTML file with annotations, by default True.
        kwargs
            Additional keyword arguments to pass to the ``cythonize`` function.
        """
        cythonize(script, exclude=exclude, nthreads=nthreads, aliases=aliases, quiet=quiet, force=force,
                  language=language, exclude_failures=exclude_failures, annotate=annotate, **kwargs)  # fmt: skip
        compiled = Path(script).with_suffix(".c")
        replaced = re.sub(f"(__PYX_EXTERN_C )?void ({'|'.join(subs)})", r'extern "C" void \2', compiled.read_text())
        compiled.write_text(replaced)
        abaqus.abaqus("make", library=str(compiled))

    def run(
        self,
        input: str,
        user: str,
        *,
        job: str = None,
        output: str = None,
        script: str = None,
        **kwargs,
    ):
        """Run Abaqus jobs.

        Parameters
        ----------
        input : str
            The path to the input file.
        user : str
            The name of the user subroutine, if it is a Cython/Pure Python script, it will be compiled
            to an object file automatically. If a companion ``.pxd`` file is found, it will be copied.
        job : str, optional
            The name of the job, by default the current directory name.
        output : str, optional
            The path to the output directory, by default the current directory.
        script : str, optional
            The Python script to run after finishing the job to post-process the results.
        kwargs
            Additional keyword arguments to pass to the ``abaqus`` command to make the object file.
        """
        # Prepare the working directory
        output = Path(output or ".").resolve()
        job = job or Path(input).stem
        user_pxd = Path(user).with_suffix(".pxd")
        if not output.exists():
            output.mkdir(parents=True)
        if not (output / Path(input).name).exists():
            shutil.copy(input, output)
        if not (output / Path(user).name).exists():
            shutil.copy(user, output)
        if user_pxd.exists() and not (output / user_pxd.name).exists():
            shutil.copy(user_pxd, output)
        if script and not (output / Path(script).name).exists():
            shutil.copy(script, output)
        os.chdir(output)

        # Compile the user subroutine if it is a Cython/Pure Python script
        if user.endswith((".pyx", ".py")):
            self.compile(Path(user).name)
            compiled = False
            for file in output.glob(f"{Path(user).stem}-*.obj"):
                user = file
                compiled = True
                break
            assert compiled, f"Failed to compile {user} to an object file."

        # Run the job
        inp = Path(input).stem
        user = Path(user).stem
        abaqus.abaqus("", job=job, input=inp, user=user, interactive=True, **kwargs)

        # Run the post-processing script
        if script:
            script = Path(script).name
            abaqus.cae(script, gui=False)


abqcy = AbqcyCLI()
